from unittest.mock import patch

from django.test import TestCase

from allianceauth.tests.auth_utils import AuthUtils

from app_utils.testdata_factories import UserMainFactory

from charlink.forms import LinkForm


class TestLinkForm(TestCase):

    @classmethod
    def setUpTestData(cls):
        cls.user = UserMainFactory()

    def test_init_no_perms(self):
        form = LinkForm(self.user)
        self.assertIn('add_character_default', form.fields)
        self.assertNotIn('allianceauth.corputils_default', form.fields)

    def test_init_with_perms(self):
        self.user = AuthUtils.add_permission_to_user_by_name('corputils.add_corpstats', self.user)
        form = LinkForm(self.user)
        self.assertIn('add_character_default', form.fields)
        self.assertIn('allianceauth.corputils_default', form.fields)

    @patch('charlink.forms.CHARLINK_IGNORE_APPS', {'allianceauth.corputils'})
    def test_init_with_perms_ignore(self):
        self.user = AuthUtils.add_permission_to_user_by_name('corputils.add_corpstats', self.user)
        form = LinkForm(self.user)
        self.assertIn('add_character_default', form.fields)
        self.assertNotIn('allianceauth.corputils_default', form.fields)



from django.test import TestCase
from django.urls import reverse

from app_utils.testdata_factories import UserMainFactory


class TestHooks(TestCase):

    @classmethod
    def setUpTestData(cls):
        cls.testuser = UserMainFactory()
        cls.html_menu = f"""
            <li>
                <a class="active" href="{reverse('charlink:index')}">
                    <i class="fas fa-link"></i> CharLink
                </a>
            </li>
        """

    def test_render_hook(self):
        self.client.force_login(self.testuser)

        response = self.client.get(reverse('charlink:index'))
        self.assertContains(response, self.html_menu, html=True)



from unittest.mock import patch

from django.test import TestCase, RequestFactory
from django.urls import reverse

from charlink.decorators import charlink


class TestCharlink(TestCase):

    @classmethod
    def setUpTestData(cls):
        cls.factory = RequestFactory()

    @patch('charlink.decorators.token_required')
    def test_ok(self, mock_token_required):
        scopes = ['esi-skills.read_skills.v1']
        session = {'charlink': {'scopes': scopes}}

        mock_token_required.return_value = lambda x: x

        @charlink
        def test_view(request):
            return 'ok'

        request = self.factory.get('/charlink/login/')
        request.session = session

        self.assertEqual(test_view(request), 'ok')
        mock_token_required.assert_called_with(scopes=scopes)

    @patch('charlink.decorators.messages.error')
    def test_missing_session(self, mock_messages_error):
        @charlink
        def test_view(request):
            return 'ok'

        request = self.factory.get('/charlink/login/')
        request.session = {}

        mock_messages_error.return_value = None

        self.assertRedirects(
            test_view(request),
            reverse('charlink:index'),
            fetch_redirect_response=False,
        )

        self.assertTrue(mock_messages_error.called)



from django.db.models import Exists, OuterRef
from django.contrib.auth.models import Permission

from moonmining.models import Owner
from moonmining import __title__, tasks
from moonmining.app_settings import MOONMINING_ADMIN_NOTIFICATIONS_ENABLED

from app_utils.messages import messages_plus
from app_utils.allianceauth import notify_admins

from allianceauth.eveonline.models import EveCorporationInfo, EveCharacter

from charlink.app_imports.utils import LoginImport, AppImport

from app_utils.allianceauth import users_with_permission


def _add_character(request, token):
    character_ownership = request.user.character_ownerships.select_related(
        "character"
    ).get(character__character_id=token.character_id)

    try:
        corporation = EveCorporationInfo.objects.get(
            corporation_id=character_ownership.character.corporation_id
        )
    except EveCorporationInfo.DoesNotExist:
        corporation = EveCorporationInfo.objects.create_corporation(
            corp_id=character_ownership.character.corporation_id
        )
        corporation.save()

    owner, _ = Owner.objects.update_or_create(
        corporation=corporation,
        defaults={"character_ownership": character_ownership},
    )
    tasks.update_owner.delay(owner.pk)
    messages_plus.success(request, f"Update of refineres started for {owner}.")
    if MOONMINING_ADMIN_NOTIFICATIONS_ENABLED:
        notify_admins(
            message=("%(corporation)s was added as new owner by %(user)s.")
            % {"corporation": owner, "user": request.user},
            title=f"{__title__}: Owner added: {owner}",
        )


def _is_character_added(character: EveCharacter):
    return Owner.objects.filter(
        character_ownership__character=character
    ).exists()


def _users_with_perms():
    return users_with_permission(
        Permission.objects.get(
            content_type__app_label='moonmining',
            codename='add_refinery_owner'
        )
    ) & users_with_permission(
        Permission.objects.get(
            content_type__app_label='moonmining',
            codename='basic_access'
        )
    )


import_app = AppImport('moonmining', [
    LoginImport(
        app_label='moonmining',
        unique_id='default',
        field_label=__title__,
        add_character=_add_character,
        scopes=Owner.esi_scopes(),
        check_permissions=lambda user: user.has_perms(["moonmining.add_refinery_owner", "moonmining.basic_access"]),
        is_character_added=_is_character_added,
        is_character_added_annotation=Exists(
            Owner.objects
            .filter(character_ownership__character_id=OuterRef('pk'))
        ),
        get_users_with_perms=_users_with_perms,
    ),
])



from dataclasses import dataclass
from typing import Callable, List

from django.http import HttpRequest
from django.db.models import Exists, QuerySet
from django import forms
from django.contrib.auth.models import User

from allianceauth.eveonline.models import EveCharacter
from esi.models import Token


@dataclass
class LoginImport:
    app_label: str
    unique_id: str
    field_label: str
    add_character: Callable[[HttpRequest, Token], None]
    scopes: List[str]
    check_permissions: Callable[[User], bool]
    is_character_added: Callable[[EveCharacter], bool]
    is_character_added_annotation: Exists
    get_users_with_perms: Callable[[], QuerySet[User]]

    def get_query_id(self):
        return f"{self.app_label}_{self.unique_id}"

    def __hash__(self) -> int:
        return hash(self.get_query_id())


@dataclass
class AppImport:
    app_label: str
    imports: List[LoginImport]

    def get_form_fields(self, user):
        return {
            import_.get_query_id(): forms.BooleanField(
                required=False,
                initial=True,
                label=import_.field_label
            )
            for import_ in self.imports
            if import_.check_permissions(user)
        }

    def get_imports_with_perms(self, user: User):
        return AppImport(
            self.app_label,
            [
                import_
                for import_ in self.imports
                if import_.check_permissions(user)
            ]
        )

    def has_any_perms(self, user: User):
        return any(import_.check_permissions(user) for import_ in self.imports)

    def get(self, unique_id: str) -> LoginImport:
        for import_ in self.imports:
            if import_.unique_id == unique_id:
                return import_

        raise KeyError(f"Import with unique_id {unique_id} not found")



import re

from django.shortcuts import render, redirect, get_object_or_404
from django.contrib.auth.decorators import login_required
from django.contrib.auth.models import User, Permission
from django.contrib import messages
from django.db.models import Q
from django.core.exceptions import PermissionDenied
from django.http import Http404

from allianceauth.services.hooks import get_extension_logger
from allianceauth.eveonline.models import EveCharacter, EveCorporationInfo
from allianceauth.authentication.decorators import permissions_required

from app_utils.allianceauth import users_with_permission

from .forms import LinkForm
from .app_imports import import_apps
from .decorators import charlink
from .app_settings import CHARLINK_IGNORE_APPS
from .utils import get_user_available_apps, get_user_linked_chars, get_visible_corps, chars_annotate_linked_apps

logger = get_extension_logger(__name__)


def get_navbar_elements(user: User):
    is_auditor = user.has_perm('charlink.view_state') or user.has_perm('charlink.view_corp') or user.has_perm('charlink.view_alliance')

    return {
        'is_auditor': is_auditor,
        'available_apps': get_user_available_apps(user) if is_auditor else [],
        'available': get_visible_corps(user) if is_auditor else [],
    }


@login_required
def index(request):
    imported_apps = import_apps()

    if request.method == 'POST':
        form = LinkForm(request.user, request.POST)
        if form.is_valid():

            scopes = set()
            selected_apps = []

            form_field_pattern = re.compile(r'^(?P<app>[\w\d\.]+)_(?P<unique_id>[a-zA-Z0-9]+)$')

            for import_code, to_import in form.cleaned_data.items():
                if to_import:
                    match = form_field_pattern.match(import_code)

                    app = match.group('app')
                    unique_id = match.group('unique_id')

                    app_import = imported_apps[app].get(unique_id)
                    scopes.update(app_import.scopes)
                    selected_apps.append((app, unique_id))

            request.session['charlink'] = {
                'scopes': list(scopes),
                'imports': selected_apps,
            }

            return redirect('charlink:login')

    else:
        form = LinkForm(request.user)

    context = {
        'form': form,
        'characters_added': get_user_linked_chars(request.user),
        **get_navbar_elements(request.user),
    }

    return render(request, 'charlink/charlink.html', context=context)


@login_required
@charlink
def login_view(request, token):
    imported_apps = import_apps()

    charlink_data = request.session.pop('charlink')

    for app, unique_id in charlink_data['imports']:
        import_ = imported_apps[app].get(unique_id)
        if app != 'add_character' and app not in CHARLINK_IGNORE_APPS and import_.check_permissions(request.user):
            try:
                import_.add_character(request, token)
            except Exception as e:
                logger.exception(e)
                messages.error(request, f"Failed to add character to {import_.field_label}")
            else:
                messages.success(request, f"Character successfully added to {import_.field_label}")

    return redirect('charlink:index')


@login_required
@permissions_required([
    'charlink.view_corp',
    'charlink.view_alliance',
    'charlink.view_state',
])
def audit(request, corp_id: int):
    corp = get_object_or_404(EveCorporationInfo, corporation_id=corp_id)
    corps = get_visible_corps(request.user)

    if not corps.filter(corporation_id=corp_id).exists():
        raise PermissionDenied('You do not have permission to view the selected corporation statistics.')

    context = {
        'selected': corp,
        **get_navbar_elements(request.user),
    }

    return render(request, 'charlink/audit.html', context=context)


@login_required
@permissions_required([
    'charlink.view_corp',
    'charlink.view_alliance',
    'charlink.view_state',
])
def search(request):
    search_string = request.GET.get('search_string', None)
    if not search_string:
        return redirect('charlink:index')

    corps = get_visible_corps(request.user)

    characters = (
        EveCharacter.objects
        .filter(
            character_name__icontains=search_string,
            corporation_id__in=corps.values('corporation_id'),
        )
        .order_by('character_name')
        .select_related('character_ownership__user__profile__main_character')
    )

    context = {
        'search_string': search_string,
        'characters': characters,
        **get_navbar_elements(request.user),
    }

    return render(request, 'charlink/search.html', context=context)


@login_required
@permissions_required([
    'charlink.view_corp',
    'charlink.view_alliance',
    'charlink.view_state',
])
def audit_user(request, user_id):
    user = get_object_or_404(User, pk=user_id)

    corps = get_visible_corps(request.user)

    if (
        not request.user.is_superuser
        and
        user != request.user
        and
        not corps
        .filter(
            corporation_id=user.profile.main_character.corporation_id
        )
        .exists()
    ):
        raise PermissionDenied('You do not have permission to view the selected user statistics.')

    context = {
        'characters_added': get_user_linked_chars(user),
        **get_navbar_elements(request.user),
    }

    return render(request, 'charlink/user_audit.html', context=context)


@login_required
@permissions_required([
    'charlink.view_corp',
    'charlink.view_alliance',
    'charlink.view_state',
])
def audit_app(request, app):
    imported_apps = import_apps()

    if app not in imported_apps:
        raise Http404()

    app_imports = imported_apps[app]

    if not app_imports.has_any_perms(request.user):
        raise PermissionDenied('You do not have permission to view the selected application statistics.')

    app_imports = app_imports.get_imports_with_perms(request.user)

    corp_ids = get_visible_corps(request.user).values('corporation_id')

    logins = {}

    for import_ in app_imports.imports:
        visible_characters = EveCharacter.objects.filter(
            (
                Q(corporation_id__in=corp_ids) |
                Q(character_ownership__user__profile__main_character__corporation_id__in=corp_ids)
            ) &
            Q(character_ownership__user__in=import_.get_users_with_perms()),
        ).select_related('character_ownership__user__profile__main_character')

        visible_characters = chars_annotate_linked_apps(
            visible_characters,
            [import_]
        ).order_by(import_.get_query_id(), 'character_name')

        logins[import_] = visible_characters

    context = {
        'logins': logins,
        'app': app,
        **get_navbar_elements(request.user),
    }

    return render(request, 'charlink/app_audit.html', context=context)



from typing import List

from django.db.models import Exists, OuterRef, Q
from django.contrib.auth.models import User

from allianceauth.authentication.models import CharacterOwnership
from allianceauth.eveonline.models import EveCharacter, EveCorporationInfo

from .app_settings import CHARLINK_IGNORE_APPS
from .app_imports import import_apps
from .app_imports.utils import LoginImport


def get_visible_corps(user: User):
    char = user.profile.main_character

    corps = EveCorporationInfo.objects.filter(
        Exists(
            CharacterOwnership.objects
            .filter(character__corporation_id=OuterRef('corporation_id'))
        )
    )

    if user.is_superuser:
        corps = corps.all()
    else:
        queries = []
        has_access = False

        if user.has_perm('charlink.view_alliance'):
            queries.append(Q(alliance__alliance_id=char.alliance_id))
            has_access = True

        if user.has_perm('charlink.view_corp') and not user.has_perm('charlink.view_alliance'):
            queries.append(Q(corporation_id=char.corporation_id))
            has_access = True

        if user.has_perm('charlink.view_state'):
            alliances = user.profile.state.member_alliances.all()
            corporations = user.profile.state.member_corporations.all()

            queries.append(
                Q(alliance__alliance_id__in=alliances.values('alliance_id')) |
                Q(id__in=corporations)
            )
            has_access = True

        if has_access:
            query = queries.pop()
            for q in queries:
                query |= q

            corps = corps.filter(query)
        else:
            corps = corps.none()

    return corps


def chars_annotate_linked_apps(characters, imports: List[LoginImport]):
    for import_ in imports:
        characters = characters.annotate(
            **{import_.get_query_id(): import_.is_character_added_annotation}
        )

    return characters


def get_user_available_apps(user: User):
    imported_apps = import_apps()

    return {
        app: imports.get_imports_with_perms(user)
        for app, imports in imported_apps.items()
        if app not in CHARLINK_IGNORE_APPS and imports.has_any_perms(user)
    }


def get_user_linked_chars(user: User):
    available_apps = get_user_available_apps(user)

    return {
        'apps': available_apps,
        'characters': chars_annotate_linked_apps(
            EveCharacter.objects.filter(character_ownership__user=user),
            [
                import_
                for imports in available_apps.values()
                for import_ in imports.get_imports_with_perms(user).imports
            ]
        )
    }



from unittest.mock import patch

from django.test import TestCase, RequestFactory

from app_utils.testdata_factories import UserMainFactory, EveCorporationInfoFactory, EveCharacterFactory
from app_utils.testing import add_character_to_user

from charlink.imports.corpstats import _add_character, _is_character_added
from charlink.app_imports import import_apps

from corpstats.models import CorpStat


class TestAddCharacter(TestCase):

    @classmethod
    def setUpTestData(cls):
        cls.user = UserMainFactory(permissions=['corpstats.add_corpstat'])
        cls.token = cls.user.token_set.first()
        cls.factory = RequestFactory()

    @patch('corpstats.models.CorpStat.update')
    def test_ok(self, mock_update):
        mock_update.return_value = None

        request = self.factory.get('/charlink/login/')

        _add_character(request, self.token)

        self.assertTrue(mock_update.called)
        self.assertTrue(_is_character_added(self.user.profile.main_character))

    @patch('allianceauth.eveonline.managers.EveCorporationManager.create_corporation', wraps=lambda corp_id: EveCorporationInfoFactory(corporation_id=corp_id))
    @patch('corpstats.models.CorpStat.update')
    def test_corp_missing(self, mock_update, mock_create_corporation):
        mock_update.return_value = None
        character = self.user.profile.main_character

        request = self.factory.get('/charlink/login/')
        character.corporation.delete()

        _add_character(request, self.token)

        self.assertTrue(mock_update.called)
        self.assertTrue(mock_create_corporation.called)
        self.assertTrue(_is_character_added(self.user.profile.main_character))


class TestIsCharacterAdded(TestCase):

    @classmethod
    def setUpTestData(cls):
        cls.user = UserMainFactory(permissions=['corputils.add_corpstats'])
        cls.character = cls.user.profile.main_character
        cls.token = cls.user.token_set.first()
        CorpStat.objects.create(token=cls.token, corp=cls.character.corporation)

    def test_ok(self):
        self.assertTrue(_is_character_added(self.character))

        newchar = EveCharacterFactory()
        add_character_to_user(self.user, newchar)

        self.assertFalse(_is_character_added(newchar))


class TestCheckPermissions(TestCase):

    @classmethod
    def setUpTestData(cls):
        cls.perm_user = UserMainFactory(permissions=['corpstats.add_corpstat'])
        cls.no_perm_user = UserMainFactory()

    def test_ok(self):
        login_import = import_apps()['corpstats'].get('default')

        self.assertTrue(login_import.check_permissions(self.perm_user))
        self.assertFalse(login_import.check_permissions(self.no_perm_user))


class TestGetUsersWithPerms(TestCase):

    @classmethod
    def setUpTestData(cls):
        cls.perm_user = UserMainFactory(permissions=['corpstats.add_corpstat'])
        cls.no_perm_user = UserMainFactory()

    def test_ok(self):
        login_import = import_apps()['corpstats'].get('default')

        users = login_import.get_users_with_perms()
        self.assertEqual(users.count(), 1)
        self.assertEqual(users.first(), self.perm_user)



from django.shortcuts import redirect
from django.contrib import messages

from esi.decorators import token_required


def charlink(func):
    def wrapper(request, *args, **kwargs):
        if 'charlink' in request.session:
            return token_required(scopes=request.session['charlink']['scopes'])(func)(request, *args, **kwargs)
        else:
            messages.error(request, 'Scopes error. Contact an administrator.')
            return redirect('charlink:index')
    return wrapper



# Generated by Django 4.0.10 on 2023-05-15 09:17

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='General',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
            ],
            options={
                'permissions': (('view_corp', 'Can view linked character of members of their corporation.'), ('view_alliance', 'Can view linked character of members of their alliance.'), ('view_state', 'Can view linked character of members of their auth state.')),
                'managed': False,
                'default_permissions': (),
            },
        ),
    ]



from importlib import import_module
from unittest.mock import patch

from django.test import TestCase

from app_utils.testdata_factories import UserMainFactory

from charlink.app_imports import import_apps
from charlink.imports.corptools import _corp_perms


class TestImportApps(TestCase):

    @patch('charlink.app_imports.import_module', wraps=import_module)
    @patch('charlink.app_imports._imported', False)
    def test_not_imported(self, mock_import_module):
        imported_apps = import_apps()
        self.assertTrue(mock_import_module.called)
        self.assertIn('add_character', imported_apps)
        self.assertIn('allianceauth.corputils', imported_apps)
        self.assertNotIn('allianceauth', imported_apps)
        self.assertNotIn('allianceauth.authentication', imported_apps)
        mock_import_module.assert_any_call('charlink.imports.allianceauth.authentication')

    @patch('charlink.app_imports.import_module', wraps=import_module)
    def test_imported(self, mock_import_module):
        import_apps()
        mock_import_module.reset_mock()
        imported_apps = import_apps()
        self.assertFalse(mock_import_module.called)
        self.assertIn('add_character', imported_apps)
        self.assertIn('allianceauth.corputils', imported_apps)
        self.assertNotIn('allianceauth', imported_apps)
        self.assertNotIn('allianceauth.authentication', imported_apps)

    def test_supported_apps_default(self):
        user = UserMainFactory()
        main_char = user.profile.main_character

        add_char = import_apps()['add_character']
        self.assertIsNone(add_char.imports[0].add_character(None, None))
        self.assertTrue(add_char.imports[0].is_character_added(main_char))
        self.assertTrue(add_char.imports[0].check_permissions(user))
        self.assertEqual(add_char.imports[0].get_users_with_perms().count(), 1)


class TestLoginImport(TestCase):

    def test_get_query_id(self):
        login_import = import_apps()['add_character'].get('default')
        self.assertEqual(login_import.get_query_id(), 'add_character_default')

    def test_hash(self):
        login_import = import_apps()['add_character'].get('default')
        self.assertEqual(hash(login_import), hash('add_character_default'))


class TestAppImport(TestCase):

    @classmethod
    def setUpTestData(cls):
        cls.user = UserMainFactory()

    def test_get_form_fields(self):
        app_import = import_apps()['add_character']
        form_fields = app_import.get_form_fields(self.user)
        self.assertEqual(len(form_fields), 1)
        self.assertIn('add_character_default', form_fields)

    def test_get_imports_with_perms(self):
        app_import = import_apps()['add_character']
        imports = app_import.get_imports_with_perms(self.user)
        self.assertEqual(len(imports.imports), 1)

        user_corp = UserMainFactory(permissions=_corp_perms)
        user_charaudit = UserMainFactory(permissions=['corptools.view_characteraudit'])
        user_both = UserMainFactory(permissions=['corptools.view_characteraudit', *_corp_perms])

        corptools_import = import_apps()['corptools']

        imports_corp = corptools_import.get_imports_with_perms(user_corp)
        self.assertEqual(len(imports_corp.imports), 1)
        self.assertEqual(imports_corp.imports[0].unique_id, 'structures')

        imports_charaudit = corptools_import.get_imports_with_perms(user_charaudit)
        self.assertEqual(len(imports_charaudit.imports), 1)
        self.assertEqual(imports_charaudit.imports[0].unique_id, 'default')

        imports_both = corptools_import.get_imports_with_perms(user_both)
        self.assertEqual(len(imports_both.imports), 2)

    def test_has_any_perms(self):
        app_import = import_apps()['add_character']
        self.assertTrue(app_import.has_any_perms(self.user))

    def test_get_ok(self):
        app_import = import_apps()['add_character']
        login_import = app_import.get('default')
        self.assertEqual(login_import.unique_id, 'default')

    def test_get_not_found(self):
        app_import = import_apps()['add_character']
        with self.assertRaises(KeyError):
            app_import.get('not_found')
